//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include "units/nCore.h"
#include "units/nWav.h"
#include "units/nShiftBlocks.h"
#include "units/nProcess.h"
#include "units/nParameter.h"
#include "units/nInitialise.h"
#include "units/nOutput.h"

int    main_argc;
char** main_argv;


int main(int argc, char* argv[])
{
	main_argc = argc;
	main_argv = argv;

	try
	{
		Check_Switches();

		if (parameters.merging)
		{
			MergeFiles();
		}
		else
		{
			if (! openWavIO())
			{
				lossyWAVerror("Error initialising wavIO unit.", 0x11);
			}

			if (Global.Codec_Block_size == 0)
			{
				lossyWAVerror("Error initialising wavIO unit.", 0x11);
			}

			initial_setup();

			if (! readNextNextCodecBlock())
			{
				lossyWAVerror("Error reading from input file.", 0x21);
			}

			Global.blocks_processed = 0;

			//==========================================================================
			// Main processing loop.
			//==========================================================================
			while (BlockData.Size.Next > 0)
			{
				Shift_Codec_Blocks();

				readNextNextCodecBlock();

				Process_This_Codec_Block();

				if (! writeNextbtrdcodecblock())
				{
					lossyWAVerror("Error writing to output file.", 0x21);
				}

				if (parameters.correction && ! writeNextcorrcodecblock())
				{
					lossyWAVerror("Error writing to correction file.", 0x22);
				}
			}

			if (! closeWavIO())
			{
				lossyWAVerror("Error closing wavIO unit.", 0x11);
			}

			write_cleanup();
		}
	}
	catch (int ret)
	{
		switch (ret)
		{
		case 0:
			return 0;
		}
	}
}
