

#include "nFFT_asm.h"

static const char UnitName [] = "nFFT_asm";


int R_Arr [ MAX_FFT_LENGTH - 1 + 1 ];
tDComplex A_Arr [ MAX_FFT_LENGTH - 1 + 1 ];
unsigned char REVERSED_BYTES [ 256 ];
int NumSamples;
int NumberOfBitsNeeded;
int MainBlockSize;
int BlockBitLen;
int PermShiftBits;
int PermShiftAdd;
double NumSamples_Reciprocal;
int Unit_Data_Packer2 [ 7 ];


int SetFFTLength( int SFL_Samples )  /* assembler */
{
	_asm
	{
		mov     Eax, SFL_Samples
		cmp     Eax, MAX_FFT_LENGTH
		ja      @@01
		cmp     Eax, 0
		jb      @@01

		//find highest set bit
		bsr     Edx, Eax
		jz      @@01
		//find lowest set bit
		bsf     Ecx, Eax
		jz      @@01
		//if highest=lowest then poweroftwo
		cmp     Edx, Ecx
		je      @@02
		//else
		@@01:     xor     Eax, Eax
		xor     Edx, Edx

		@@02:     mov     Ecx, MAX_FFT_BIT_LENGTH
		sub     Ecx, Edx
		mov     PermShiftBits, Ecx
		mov     Ecx, DWORD ptr PowersOf.TwoInt[Ecx* INT64_SIZE]
		mov     PermShiftAdd, Ecx

		mov     NumberOfBitsNeeded, Edx
		mov     NumSamples, Eax
		fld     qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE-Edx* DOUBLE_SIZE]
		fstp    NumSamples_Reciprocal

	}
}


void Shuffle_In_Place_Complex( )  /* assembler */
{
	_asm
	{

		mov     Esi, NumSamples
		dec     Esi
		mov     Edx, PermShiftAdd
		xor     Edi, Edi
		xor     Ecx, Ecx
		jmp     @@S2
		nop

		@@S1:
		mov     Eax, DWORD ptr R_Arr[Edi* INTEGER_SIZE]


		cmp     Eax, Ecx
		jle     @@S2

		fld     qword ptr FFT_Array[Ecx]
		fld     qword ptr FFT_Array[Ecx+DOUBLE_SIZE]

		fld     qword ptr FFT_Array[Eax]
		fld     qword ptr FFT_Array[Eax+DOUBLE_SIZE]

		fstp    qword ptr FFT_Array[Ecx+DOUBLE_SIZE]
		fstp    qword ptr FFT_Array[Ecx]

		fstp    qword ptr FFT_Array[Eax+DOUBLE_SIZE]
		fstp    qword ptr FFT_Array[Eax]



		@@S2:       Add     Edi, Edx
		Add     Ecx, COMPLEX_SIZE
		sub     Esi, 1
		jg      @@S1

	}
}


void FFT_DIT_Complex( )  /* assembler */
_asm
{
	push    Ebx
	push    Esi
	push    Edi
	push    Ebp

	call    Shuffle_In_Place_Complex

	mov     BlockBitLen, 0


	@@R4:     mov     Eax, BlockBitLen
	Add     Eax, 2
	cmp     Eax, NumberOfBitsNeeded
	ja      @@R2

	call    FFT_Radix_Four_DIT
	jmp     @@R4

	@@R2:     mov     Eax, BlockBitLen
	Add     Eax, 1
	cmp     Eax, NumberOfBitsNeeded
	ja      @@End

	call    FFT_Radix_Two_DIT
	jmp     @@R2

	@@End:    pop     Ebp
	pop     Edi
	pop     Esi
	pop     Ebx
	//============================================================================

}
void FFT_Radix_Four_DIT( )  /* assembler */
_asm
{

	mov     BlockBitLen, Eax

	mov     Ecx, DWORD ptr PowersOf.TwoInt[Eax* INT64_SIZE]
	mov     MainBlockSize, Ecx
	mov     Ecx, DWORD ptr PowersOf.TwoInt[Eax*INT64_SIZE-2*INT64_SIZE]
	mov     Stride, Ecx

	mov     Eax, MAX_FFT_BIT_LENGTH
	sub     Eax, BlockBitLen
	mov     ShiftBits, Eax

	mov     i, 0
	//================================================================
	@@FC0:      mov     Eax, i
	shl     Eax, COMPLEX_SHIFT
	mov     J, Eax
	mov     n, 0
	//==============================================================
	@@FC1:        mov     Esi, n
	mov     Ecx, ShiftBits
	Add     Ecx, COMPLEX_SHIFT
	shl     Esi, cl

	mov     Edi, Esi
	Add     Edi, Esi

	mov     Edx, Stride
	shl     Edx, COMPLEX_SHIFT

	mov     Eax, J
	mov     Ebx, Eax
	Add     Ebx, Edx //K
	mov     Ecx, Ebx
	Add     Ecx, Edx //L
	Add     Edx, Ecx //M

	//============================================================
	fld     qword ptr FFT_Array[Edx]
	fld     qword ptr FFT_Array[Edx+DOUBLE_SIZE]
	fld     qword ptr A_Arr[Edi+Esi+DOUBLE_SIZE]
	fld     qword ptr A_Arr[Edi+Esi]

	fld     st
	fmul    st, st( 3 )
	fld     st( 2 )
	fmul    st, st( 5 )
	fadd
	fxch    st( 4 )

	fmul
	fxch    st( 2 )
	fmul
	fsub

	fld     qword ptr FFT_Array[Ecx]
	fld     qword ptr FFT_Array[Ecx+DOUBLE_SIZE]
	fld     qword ptr A_Arr[Esi+DOUBLE_SIZE]
	fld     qword ptr A_Arr[Esi]

	fld     st
	fmul    st, st( 3 )
	fld     st( 2 )
	fmul    st, st( 5 )
	fadd
	fxch    st( 4 )

	fmul
	fxch    st( 2 )
	fmul
	fsub

	fxch    st( 3 )

	fld     st( 3 )
	fadd    st, st( 3 )
	fstp    Y1.Re
	fld     st( 1 )
	fadd    st, st( 1 )
	fstp    Y1.Im

	fsubr
	fstp    Y3.Re
	fsub
	fstp    Y3.Im
	//============================================================
	fld     qword ptr FFT_Array[Ebx]
	fld     qword ptr FFT_Array[Ebx+DOUBLE_SIZE]
	fld     qword ptr A_Arr[Edi+DOUBLE_SIZE]
	fld     qword ptr A_Arr[Edi]

	fld     st
	fmul    st, st( 3 )
	fld     st( 2 )
	fmul    st, st( 5 )
	fadd
	fxch    st( 4 )

	fmul
	fxch    st( 2 )
	fmul
	fsub

	fld     qword ptr FFT_Array[Eax]
	fld     st
	fadd    st, st( 2 )
	fstp    Y0.Re
	fsubr
	fstp    Y2.Re

	fld     qword ptr FFT_Array[Eax+DOUBLE_SIZE]
	fld     st
	fadd    st, st( 2 )
	fstp    Y0.Im
	fsubr
	fstp    Y2.Im
	//============================================================
	fld     Y0.Re
	fld     Y1.Re
	fld     Y0.Im
	fld     Y1.Im

	fld     st
	fadd    st, st( 2 )
	fstp    qword ptr FFT_Array[Eax+DOUBLE_SIZE]
	fsub
	fstp    qword ptr FFT_Array[Ecx+DOUBLE_SIZE]

	fld     st
	fadd    st, st( 2 )
	fstp    qword ptr FFT_Array[Eax]
	fsub
	fstp    qword ptr FFT_Array[Ecx]
	//============================================================
	fld     Y3.Im
	fld     Y2.Im
	fld     Y3.Re
	fld     Y2.Re

	fld     st
	fsub    st, st( 2 )
	fstp    qword ptr FFT_Array[Ebx]
	fadd
	fstp    qword ptr FFT_Array[Edx]

	fld     st
	fsub    st, st( 2 )
	fstp    qword ptr FFT_Array[Ebx+DOUBLE_SIZE]
	fadd
	fstp    qword ptr FFT_Array[Edx+DOUBLE_SIZE]
	//============================================================
	Add    J, COMPLEX_SIZE
	mov    Eax, n
	Add    Eax, 1
	mov    n, Eax
	cmp    Eax, Stride
	jl     @@FC1
	//==============================================================
	mov    Eax, i
	Add    Eax, MainBlockSize
	mov    i, Eax
	cmp    Eax, NumSamples
	jl     @@FC0
	//==========================================================================

}
int MainBlockSize;
int Stride;
int ShiftBits;
int i, J, n;
tDComplex Y0, Y1, Y2, Y3;
void FFT_Radix_Two_DIT( )  /* assembler */
{
	_asm
	{

		mov     BlockBitLen, Eax

		mov     Ecx, DWORD ptr PowersOf.TwoInt[Eax* INT64_SIZE]
		mov     MainBlockSize, Ecx
		shl     Ecx, COMPLEX_SHIFT-1



		@@FC3:      xor     Ebp, Ebp                                      //i
		mov     Edi, MAX_FFT_BIT_LENGTH+COMPLEX_SHIFT
		sub     Edi, BlockBitLen
		mov     Edi, DWORD ptr PowersOf.TwoInt[Edi* INT64_SIZE]



		@@FC4:        xor     Esi, Esi                                    //n
		xor     Eax, Eax

		@@FC5:          mov     Ebx, Ebp
		shl     Ebx, COMPLEX_SHIFT
		Add     Ebx, Esi                                  //j
		mov     Edx, Ebx
		Add     Edx, Ecx                                  //k:=j+blockend

		// load FFT_Array[k] into FPU
		fld     qword ptr FFT_Array[Edx]                  // D1.Re
		fld     qword ptr FFT_Array[Edx+DOUBLE_SIZE]      // D1.Im
		fld     qword ptr A_Arr[Eax+DOUBLE_SIZE]          // W1.Im
		fld     qword ptr A_Arr[Eax]                      // W1.Re

		fld     st                                        // W1.Re
		fmul    st, st( 3 )                               // W1.Re * D1.Im
		fld     st( 2 )                                   // W1.Im
		fmul    st, st( 5 )                               // W1.Im * D1.Re
		fadd                                              // W1.Re * D1.Im + W1.Im * D1.Re
		fxch    st( 4 )

		fmul                                              // D1.Re * W1.Re
		fxch    st( 2 )                                   // D1.Im
		fmul                                              // W1.Im * D1.Im
		fsub                                              // D1.Re * W1.Re - W1.Im * D1.Im

		fld     qword ptr FFT_Array[Ebx]                  // FFT_Array[j].re
		fld     st
		fsub    st, st( 2 )
		fstp    qword ptr FFT_Array[Edx]                  // FFT_Array[k].re:=FFT_Array[j].re-t.re
		fadd
		fstp    qword ptr FFT_Array[Ebx]                  // FFT_Array[j].re:=FFT_Array[j].re+t.re

		fld     qword ptr FFT_Array[Ebx+DOUBLE_SIZE]      // FFT_Array[j].im
		fld     st
		fsub    st, st( 2 )
		fstp    qword ptr FFT_Array[Edx+DOUBLE_SIZE]      // FFT_Array[k].im:=FFT_Array[j].im-t.im
		fadd
		fstp    qword ptr FFT_Array[Ebx+DOUBLE_SIZE]      // FFT_Array[j].im:=FFT_Array[j].im+t.im

		Add     Eax, Edi
		Add     Esi, COMPLEX_SIZE                         // n:=n+1
		cmp     Esi, Ecx
		jl      @@FC5

		Add     Ebp, MainBlockSize                          // fft_i:=fft_i+MainBlockSize
		cmp     Ebp, NumSamples
		jl      @@FC4
		//==========================================================================

	}
}


void FFT_DIT_Real( )  /* assembler */
{
	_asm
	{

		push    Ebx
		push    Ebp
		push    Edi
		push    Esi

		call    FFT_DIT_Complex

		mov     Ecx, PermShiftBits
		dec     Ecx

		mov     Ebx, NumSamples
		shl     Ebx, COMPLEX_SHIFT

		mov     Edi, Ebx
		shr     Edi, 1

		fld     SINGLE_HALF                             //0.5        #0#


		@@FR1:      mov     Esi, Edi
		shl     Esi, cl

		mov     Eax, Ebx
		sub     Eax, Edi
		//  -
		fld     qword ptr FFT_Array[Edi]              //temp1.re   #1#
		fld     qword ptr FFT_Array[Eax]              //temp2.re   #2#
		fld     qword ptr FFT_Array[Edi+DOUBLE_SIZE]  //temp1.im   #3#
		fld     qword ptr FFT_Array[Eax+DOUBLE_SIZE]  //temp2.im   #4#

		fld     st( 3 )                               //temp1.re    5
		fsub    st, st( 3 )                           //temp2.re    5
		fmul    qword ptr A_Arr[Esi]                  //ar          5
		fld     st( 2 )                               //temp1.im    6
		fadd    st, st( 2 )                           //temp2.im    6
		fmul    qword ptr A_Arr[Esi+DOUBLE_SIZE]      //ai          6
		fsub                                          //T.im       #5#

		fld     st( 2 )                               //temp1.im    6
		fsub    st, st( 2 )                           //temp2.im    6
		fsub    st, st( 1 )                           //T.Im        6
		fmul    st, st( 6 )                           //0.5         6
		fst     qword ptr FFT_Array[Edi+DOUBLE_SIZE]            //  6
		fchs
		fst     qword ptr FFT_Array[Ebx+Eax+DOUBLE_SIZE]
		fchs
		fadd                                                    //  5
		fst     qword ptr FFT_Array[Ebx+Edi+DOUBLE_SIZE]
		fchs
		fstp    qword ptr FFT_Array[Eax+DOUBLE_SIZE]            // #4#

		fadd                                                    //  3
		fmul    qword ptr A_Arr[Esi]                  //ar          3
		fld     st( 2 )                               //temp1.re    4
		fsub    st, st( 2 )                           //temp2.re    4
		fmul    qword ptr A_Arr[Esi+DOUBLE_SIZE]      //ai          4
		fadd                                          //T.re       #3#

		fxch    st( 2 )                                         //  3
		fadd                                                    //  2
		fadd    st, st( 1 )                                     //  2
		fmul    st, st( 2 )                           //0.5         2
		fst     qword ptr FFT_Array[Edi]                        //  2
		fst     qword ptr FFT_Array[Ebx+Eax]                        //  2
		fsubr                                                   //  1
		fst     qword ptr FFT_Array[Ebx+Edi]
		fstp    qword ptr FFT_Array[Eax]                        // #0#

		sub     Edi, COMPLEX_SIZE
		jg      @@FR1

		fstp    st                                                // #-#

		fld     qword ptr FFT_Array[0]
		fld     qword ptr FFT_Array[DOUBLE_SIZE]

		fld     st( 1 )
		fadd    st, st( 1 )
		fstp    qword ptr FFT_Array[0]
		fsub
		fstp    qword ptr FFT_Array[Ebx]

		fldz
		fst     qword ptr FFT_Array[DOUBLE_SIZE]
		fstp    qword ptr FFT_Array[Ebx+DOUBLE_SIZE]

		pop     Esi
		pop     Edi
		pop     Ebp
		pop     Ebx

	}
}


void IFFT_DIT_Complex( )  /* assembler */
{
	_asm
	{

		push    Ebx
		push    Esi
		push    Edi
		push    Ebp

		call    Shuffle_In_Place_Complex

		mov     Ecx, 1                                          // blockend
		mov     BlockBitLen, 1                                  // blockbitlen
		mov     MainBlockSize, 2                                // MainBlockSize


		@@FC3:      xor     Ebp, Ebp                                      //i
		mov     Edi, MAX_FFT_BIT_LENGTH
		sub     Edi, BlockBitLen
		Add     Edi, COMPLEX_SHIFT
		mov     Edi, DWORD ptr PowersOf.TwoInt[Edi* INT64_SIZE]



		@@FC4:        xor     Esi, Esi                                     //n
		xor     Eax, Eax

		@@FC5:          fld     qword ptr A_Arr[Eax+DOUBLE_SIZE]          //a.im:=a_ptr[blockbitlen]^[n].Im
		fchs
		fld     qword ptr A_Arr[Eax]                      //a.re:=a_ptr[blockbitlen]^[n].Re

		mov     Ebx, Esi                                  //j
		Add     Ebx, Ebp
		mov     Edx, Ebx
		Add     Edx, Ecx                                  //k:=j+blockend

		shl     Ebx, COMPLEX_SHIFT
		shl     Edx, COMPLEX_SHIFT

		// load FFT_Array[k] into FPU
		fld     qword ptr FFT_Array[Edx+DOUBLE_SIZE]      // FFT_Array[k].im
		fld     qword ptr FFT_Array[Edx]                  // FFT_Array[k].re

		// T.re := A.re * FFT_Array.Complex[K].Re - A.im * FFT_Array.Complex[K].Im;
		fld     st( 0 )
		fmul    st, st( 3 )
		fld     st( 2 )
		fmul    st, st( 5 )
		fsub                                              // t.re
		fxch    st( 4 )

		// T.im := A.re * FFT_Array.Complex[K].Im + A.im * FFT_Array.Complex[K].Re;
		fmul
		fxch    st( 2 )
		fmul
		fadd                                              // t.im

		fld     qword ptr FFT_Array[Ebx+DOUBLE_SIZE]      // FFT_Array[j].im
		fld     st
		fsub    st, st( 2 )
		fstp    qword ptr FFT_Array[Edx+DOUBLE_SIZE]      // FFT_Array[k].im:=FFT_Array[j].im-t.im
		fadd
		fstp    qword ptr FFT_Array[Ebx+DOUBLE_SIZE]      // FFT_Array[j].im:=FFT_Array[j].im+t.im

		fld     qword ptr FFT_Array[Ebx]                  // FFT_Array[j].re
		fld     st
		fsub    st, st( 2 )
		fstp    qword ptr FFT_Array[Edx]                  // FFT_Array[k].re:=FFT_Array[j].re-t.re
		fadd
		fstp    qword ptr FFT_Array[Ebx]                  // FFT_Array[j].re:=FFT_Array[j].re+t.re

		Add     Esi, 1                                    // n:=n+1
		Add     Eax, Edi
		cmp     Esi, Ecx
		jl      @@FC5

		Add     Ebp, MainBlockSize                          // fft_i:=fft_i+MainBlockSize
		cmp     Ebp, NumSamples
		jl      @@FC4

		shl     Ecx, 1                                        // blockend:=2*blockend;
		shl     MainBlockSize, 1                              // MainBlockSize:=2*MainBlockSize;

		Add     BlockBitLen, 1
		mov     Eax, BlockBitLen
		cmp     Eax, NumberOfBitsNeeded
		jle     @@FC3

		mov     Ecx, NumSamples
		shl     Ecx, COMPLEX_SHIFT

		pop     Ebp
		pop     Edi
		pop     Esi
		pop     Ebx

	}
}


void nfft_double_init( )  /* assembler */
{
	_asm
	{
		push   Ebx

		sub    Ebx, Ebx
		sub    Ecx, Ecx
		sub    Edx, Edx

		@@q1:     mov    al, byte ptr REVERSED_NIBBLES[Ecx]
		shl    al, 4
		or     al, byte ptr REVERSED_NIBBLES[Edx]

		mov    bl, dl
		shl    bl, 4
		or     bl, cl
		mov    byte ptr REVERSED_BYTES[Ebx], al

		Add    Edx, 1
		cmp    Edx, $10
		jl     @@q1

		sub    Edx, Edx
		Add    Ecx, 1
		cmp    Ecx, $10
		jl     @@q1

		pop    Ebx
		//========================================================================

		//========================================================================
		// Fill ReversedBits lookup table (shifted to take into account
		// COMPLEX_SIZE) for maximum FFT length (shift index to suit).
		//========================================================================
		@@01:     mov     Edx, MAX_FFT_LENGTH-1

		@@02:       mov     Eax, Edx
		sub     Ecx, Ecx

		mov     cl, al
		mov     al, byte ptr REVERSED_BYTES[Ecx]
		mov     cl, ah
		mov     ah, byte ptr REVERSED_BYTES[Ecx]

		bswap   Eax

		mov     cl, al
		mov     al, byte ptr REVERSED_BYTES[Ecx]
		mov     cl, ah
		mov     ah, byte ptr REVERSED_BYTES[Ecx]

		shr     Eax, 32-MAX_FFT_BIT_LENGTH-COMPLEX_SHIFT
		mov     DWORD ptr R_Arr[Edx* INTEGER_SIZE], Eax

		sub     Edx, 1
		jge     @@02
		//========================================================================


		//========================================================================
		// Calculate A lookup table for maximum FFT length (shift index to suit).
		//========================================================================
		@@04:     fldpi
		fmul      qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE-( MAX_FFT_BIT_LENGTH*DOUBLE_SIZE-DOUBLE_SIZE )]
		fchs        // -(2 x pi)/(2^MAX_FFT_BIT_LENGTH)

		fldz

		mov     Ecx, MAX_FFT_LENGTH* COMPLEX_SIZE
		sub     Eax, Eax

		@@05:         fld     st
		fsincos
		fstp    qword ptr A_Arr[Eax]                //re
		fstp    qword ptr A_Arr[Eax+DOUBLE_SIZE]    //im

		fadd    st, st( 1 )
		Add     Eax, COMPLEX_SIZE
		cmp     Eax, Ecx
		jl      @@05

		@@06:     //end of the loop
		fstp    st
		fstp    st
		//========================================================================

	}
}
// initialization: todo
Unit_Data_Packer2[0] = 0;
write_unitname_and_datapointers( UnitName, & Unit_Data_Packer1, & Unit_Data_Packer2, sizeof( Unit_Data_Packer2 ) );

if ( !( FFTW_Initialised( ) ) )
{
	nfft_double_init( );
}

// finalization: todo Write ( Erroutput, UnitName, CRLF );
