// --------------------------------------------------------------
//  Delphi interface to the FFTW library -- FFTW Version 3.0.1.
//  Note that this interface is incomplete. Additional function
//  interface entries may be added in an anologous manner, see
//  FFTW for more details.
//
//  Last modified 22/DEC/03
//  Written by Mark G. Beckett (g.beckett@epcc.ed.ac.uk
// --------------------------------------------------------------
//  Modified 07/JUN/2009
//  by Nick Currie (lossyWAV at yahoo dot com)
//  Now dynamically links rather than statically. Required to
//  stop program crashing if libFFTW3-3.DLL is not available.
// --------------------------------------------------------------
//  Modified 14/APR/2010
//  by Nick Currie (lossyWAV at hotmail <dot> co <dot> uk
// --------------------------------------------------------------
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
// --------------------------------------------------------------

#include "FFTW_interface.h"

static const char UnitName [] = "FFTW_Interface";

FFTW_Rec FFTW;

HMODULE FFTW_DLL_Handle;
bool FFTW_DLL_Loaded;


bool Check_Initialised(FFTW_Rec& FFTW_Record)
{
	return ((& FFTW_Record.Plan_DFT_r2c_1d != NULL) && (& FFTW_Record.Plan_DFT_1d != NULL) && (& FFTW_Record.Execute != NULL) && (& FFTW_Record.Destroy_Plan != NULL));
}


bool FFTW_Initialised()
{
	return FFTW_DLL_Loaded && (FFTW_DLL_Handle != 0);
}


void FFTW_Initialise()
{
#if defined(__GNUC__)
    FFTW_DLL_Handle = LoadLibrary("libfftw3-3.dll");
#else
    FFTW_DLL_Handle = LoadLibrary(L"libfftw3-3.dll");
#endif
	if (FFTW_DLL_Handle != 0)
	{
		FFTW.Plan_DFT_r2c_1d = (void*(*)(int,double*,double*,unsigned int)) GetProcAddress(FFTW_DLL_Handle, "fftw_plan_dft_r2c_1d");
		FFTW.Plan_DFT_1d = (void*(*)(int,double*,double*,int,unsigned int)) GetProcAddress(FFTW_DLL_Handle, "fftw_plan_dft_1d");
		FFTW.Execute = (void(*)(void*)) GetProcAddress(FFTW_DLL_Handle, "fftw_execute");
		FFTW.Destroy_Plan = (void(*)(void*)) GetProcAddress(FFTW_DLL_Handle, "fftw_destroy_plan");
	}

	FFTW_DLL_Loaded = Check_Initialised(FFTW) && (FFTW_DLL_Handle != 0);
}


bool CHECK_FFTW_DLL_Loaded()
{
	return FFTW_DLL_Loaded;
}


void FFTW_Cleanup()
{
	int fc_i;

	if (FFTW_DLL_Handle != 0)
	{
		/*with FFTW do*/
		{
			for (fc_i = 1; fc_i != MAX_FFT_BIT_LENGTH; ++fc_i)
			{
				if (FFTW.Plans[fc_i] != NULL)
				{
					FFTW.Destroy_Plan(FFTW.Plans[fc_i]);
				}
			}

			FFTW.Plan_DFT_r2c_1d = NULL;
			FFTW.Plan_DFT_1d = NULL;
			FFTW.Execute = NULL;
			FFTW.Destroy_Plan = NULL;
		}
		FreeLibrary(FFTW_DLL_Handle);
		FFTW_DLL_Handle = 0;
		FFTW_DLL_Loaded = false;
	}
}

// initialization: todo
static struct Init
{
	Init() {
		// initialization: todo
		FFTW_Initialise();
	}
	~Init() {
		// finalization: todo FFTW_Cleanup ( );
		//Write(Erroutput, UnitName, CRLF);
	}
} init;
