//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include "nMaths.h"
#include "nCore.h"
#include "nFillFFT.h"

static const char UnitName [] = "nFillFFT";

// Global
FillFFT_type FillFFT;
double* window_function[ MAX_FFT_BIT_LENGTH + 1 ];


struct SavedBlocks_type
{
	MultiChannelCodecBlock BTRD;
	MultiChannelCodecBlock CORR;
}
SavedBlocks;

struct fill_fft_lookup_type
{
	int block;
	int Offset;
}
fill_fft_lookup [ MAX_BLOCK_SIZE * 4 ];


void nFillFFT_Init()
{
	int ff_i;

	for (ff_i = 0; ff_i < Global.Codec_Block_size * 4;  ++ff_i)
	{
		fill_fft_lookup[ff_i].block = (ff_i / Global.Codec_Block_size);
		fill_fft_lookup[ff_i].Offset = (ff_i % Global.Codec_Block_size);
	}

	FillFFT.Codec_Block_fill_offset = Global.Codec_Block_size * 2;
}

double Apply_Window_Function(double Filled_Average)
{
	int ff_i;
	double ff_l, ff_m;
	ff_l = 0;
	ff_i = 0;

	while (ff_i < Global.FFT.length)
	{
		ff_m = FFT_Array.DReal[ff_i] - Filled_Average;
		ff_l = ff_l + nsqr(ff_m);
		FFT_Array.DReal[ff_i] = ff_m * window_function[Global.FFT.bit_length][ff_i];
		++ ff_i;
	}

	return nlog2(ff_l * Global.FFT.length_recip) * SINGLE_HALF;
}


double FillFFT_Input_From_Data()
{
	int ff_i, ff_j, ff_n;
	double ff_k, ff_m;
	ff_k = 0;
	ff_n = FillFFT.Codec_Block_fill_offset + FillFFT.actual_analysis_block_start;
	ff_i = 0;

	while (ff_i < Global.FFT.length)
	{
		ff_j = ff_i + ff_n;
		ff_m = BlockData.Pointer[ fill_fft_lookup[ff_j].block ][ Global.channel ][ fill_fft_lookup[ff_j].Offset ] * settings.scaling_factor;
		ff_k = ff_k + ff_m;
		FFT_Array.DReal[ff_i] = ff_m;
		++ ff_i;
	}

	return Apply_Window_Function(ff_k * Global.FFT.length_recip * settings.dccorrect_multiplier);
}


double FillFFT_Input_From_BTRD()
{
	int ff_i, ff_j;
	double ff_k, ff_m;
	ff_k = 0;
	ff_i = 0;

	while (ff_i < Global.FFT.length)
	{
		ff_j = ff_i + FillFFT.actual_analysis_block_start;

		if (ff_j < 0)
		{
			ff_m = SavedBlocks.BTRD[Global.channel][ff_j + Global.Codec_Block_size];
		}
		else if (ff_j > Global.Codec_Block_size)
		{
			ff_m = BlockData.Pointer[3][Global.channel][ff_j - Global.Codec_Block_size] * settings.scaling_factor;
		}
		else
		{
			ff_m = BlockData.BTRD[Global.channel][ff_j];
		}

		ff_k = ff_k + ff_m;
		FFT_Array.DReal[ff_i] = ff_m;
		++ ff_i;
	}

	return Apply_Window_Function(ff_k * Global.FFT.length_recip * settings.dccorrect_multiplier);
}


double FillFFT_Input_From_CORR()
{
	int ff_i, ff_j;
	double ff_k, ff_m;
	ff_k = 0;
	ff_i = 0;

	while (ff_i < Global.FFT.length)
	{
		ff_j = ff_i + FillFFT.actual_analysis_block_start;

		if (ff_j < 0)
		{
			ff_m = SavedBlocks.CORR[Global.channel][ff_j + Global.Codec_Block_size];
		}
		else if (ff_j > Global.Codec_Block_size)
		{
			ff_m = BlockData.Pointer[3][Global.channel][ff_j - Global.Codec_Block_size] * settings.scaling_factor;
		}
		else
		{
			ff_m = BlockData.CORR[Global.channel][ff_j];
		}

		ff_k = ff_k + ff_m;
		FFT_Array.DReal[ff_i] = ff_m;
		++ ff_i;
	}

	return Apply_Window_Function(ff_k * Global.FFT.length_recip * settings.dccorrect_multiplier);
}


void Fill_Audio_Safes()
{
	size_t n = Global.channels * Global.Codec_Block_size;

	std::copy(&BlockData.BTRD[0][0], &BlockData.BTRD[0][0] + n, &SavedBlocks.BTRD[0][0]);
	std::copy(&BlockData.CORR[0][0], &BlockData.CORR[0][0] + n, &SavedBlocks.CORR[0][0]);
/*
	int ff_i, ff_j;

	for (ff_j = 0; ff_j < Global.channels; ++ff_j)
	{
		for (ff_i = 0; ff_i < Global.Codec_Block_size; ++ff_i)
		{
			SavedBlocks.BTRD[ff_j][ff_i] = BlockData.BTRD[ff_j][ff_i];
			SavedBlocks.CORR[ff_j][ff_i] = BlockData.CORR[ff_j][ff_i];
		}
	}
*/
}


void Window_Function_Lookup_Init()
{
	int ff_i, ff_j;
	// Generate window_function lookup tables
	for (ff_j = 0; ff_j <= MAX_FFT_BIT_LENGTH; ++ff_j)
	{
		window_function[ff_j] = new double[ size_t(PowersOf.TwoInt[ff_j + 3]) ];

		for (ff_i = 0; ff_i <= PowersOf.TwoM1[ff_j]; ++ff_i)
		{
			window_function[ff_j][ff_i] = Hann(ff_i, ff_j);
		}
	}

	FillFFT.Window_Function_RMS = nlog2(3.0 / 8.0) / 2.0 * db_lg2x20;  // == RMS of Hann Window x 20 x log10(2).
}


void window_function_lookup_final()
{
	int ff_i;

	for (ff_i = 0; ff_i <= MAX_FFT_BIT_LENGTH; ++ff_i)
	{
		delete[] window_function[ff_i];
	}
}

static struct Init
{
	Init() {
		// initialization: todo
		Window_Function_Lookup_Init();
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
		window_function_lookup_final();
	}
} init;
