//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include <cmath>
#include "nMaths.h"
#include "nCore.h"

static const char UnitName [] = "nMaths";

int temp_integer = 0;


double Hann(int Index, int fft_bit_length)
{
	return SINGLE_HALF * (SINGLE_ONE - std::cos(Index * TwoPi * PowersOf.Two[-fft_bit_length]));
}


double nround10(double nr_v, int nr_d)
{
	if (abs(nr_d) < 307)
	{
		return nround(nr_v * PowersOf.Ten[nr_d]) * PowersOf.Ten[-nr_d];
	}

	return nr_v;
}


double npower2(double fp_2v)
{
	return std::pow(2.0, fp_2v);
}


double nlog2(double fl_2v)
{
	static const double log_2 = std::log(2.0); // computed only once

	if (fl_2v > 0)
	{
		//return std::log2(fl_2v); // c++11
		return std::log(fl_2v) / log_2;
	}

	return 0.0;
}


double nlog10(double fl_10v)
{
	if (fl_10v > 0.0)
	{
		return std::log10(fl_10v);
	}

	return 0.0;
}


double nroot(double fr_v)
{
	if (fr_v > 0.0)
	{
		return std::sqrt(fr_v);
	}

	return 0.0;
}
// initialization: todo
//temp_integer = 0;
// finalization: todo Write ( Erroutput, UnitName, CRLF );
