//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include <iostream>
#include <cmath>

#if defined(_MSC_VER)
#include <intrin.h>
#pragma intrinsic(_BitScanForward)
#pragma intrinsic(_BitScanReverse)
#endif

#include "nOutput.h"
#include "nCore.h"
#include "nMaths.h"
#include "nSpreading.h"
#include "fftw_interface.h"
#include "nSGNS.h"
#include "nParameter.h"

static const char UnitName [] = "nOutput";

// Global
unsigned char* bit_removal_history;

const char hyphen_string[256] = "---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------";
const char bits_filled[256]   = "OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO000000000000OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO000000000000OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO000000000000OOOOOO";
const char bits_empty[256]    = "...............................................................................................................................................................................................................................................................";
const char space_string[256]  = "                                                                                                                                                                                                                                                               ";


const int  results_offset [ MAX_CHANNELS ] =
{
	0 * 34 * INTEGER_SIZE,
	1 * 34 * INTEGER_SIZE,
	2 * 34 * INTEGER_SIZE,
	3 * 34 * INTEGER_SIZE,
	4 * 34 * INTEGER_SIZE,
	5 * 34 * INTEGER_SIZE,
	6 * 34 * INTEGER_SIZE,
	7 * 34 * INTEGER_SIZE
};


int DATA_block_value;
int BTRD_block_value;
int CORR_block_value;

int DATA_block_lsb [ MAX_CHANNELS ][ 34 ];
int DATA_block_msb [ MAX_CHANNELS ][ 34 ];
int DATA_sample_lsb [ MAX_CHANNELS ][ 34 ];

int DATA_sample_msb [ MAX_CHANNELS ][ 34 ];
int BTRD_block_lsb [ MAX_CHANNELS ][ 34 ];
int BTRD_block_msb [ MAX_CHANNELS ][ 34 ];

int BTRD_sample_lsb [ MAX_CHANNELS ][ 34 ];
int BTRD_sample_msb [ MAX_CHANNELS ][ 34 ];
int CORR_block_lsb [ MAX_CHANNELS ][ 34 ];

int CORR_block_msb [ MAX_CHANNELS ][ 34 ];
int CORR_sample_lsb [ MAX_CHANNELS ][ 34 ];
int CORR_sample_msb [ MAX_CHANNELS ][ 34 ];

std::string head_bar;
std::string top_bar;
std::string mid_bar;
int Display_Width;
int bar_length;
std::string Titles [ 16 ];
std::string Header;


void Make_Bars(int MB_Width, int MB_Item, int MB_Value, int MB_Count, int MB_Summary = 0)
{
	int mb_i;

	if (MB_Summary > 0)
	{
		bar_length = ((MB_Width - (MB_Item + 2) - (MB_Value + 1) * MB_Count - (MB_Summary + 1)) / MB_Count) - 1;
	}
	else
	{
		bar_length = ((MB_Width - (MB_Item + 2) - (MB_Value + 1) * MB_Count) / MB_Count) - 1;
	}

	top_bar = "+" + std::string(hyphen_string, MB_Item) + '+';
	mid_bar = top_bar;
	head_bar = top_bar + std::string(hyphen_string, Display_Width - top_bar.length() - 1) + "+";

	if (bar_length > 0)
		for (mb_i = 1; mb_i <= MB_Count;  ++mb_i)
		{
			top_bar += std::string(hyphen_string, MB_Value + 1 + bar_length) + "+";
			mid_bar += std::string(hyphen_string, MB_Value) + "+" + std::string(hyphen_string, bar_length) + "+";
		}
	else
		for (mb_i = 1; mb_i <= MB_Count;  ++mb_i)
		{
			top_bar += std::string(hyphen_string, MB_Value) + "+";
			mid_bar += std::string(hyphen_string, MB_Value) + "+";
		}

	if (MB_Summary > 0)
	{
		top_bar += std::string(hyphen_string, MB_Summary) + "+";
		mid_bar += std::string(hyphen_string, MB_Summary) + "+";
	}

	if (bar_length > 0)
	{
		for (mb_i = 0; mb_i < MB_Count;  ++mb_i)
		{
			Titles[mb_i] += std::string(space_string, bar_length + 2 + MB_Value - sizeof /*Length*/ (Titles[mb_i]));
		}

		Header += std::string(space_string, MB_Count * (bar_length + 2 + MB_Value) - Header.length());
	}
}


void remove_bits_detailed_output(std::ostream& ToOutput)
{
	int rb_i, rb_j, rb_k, rb_l;
	int ss_k;
	double btr_row_tot;
	int channel;
	std::string time_string;
    ToOutput << CRLF << "Detailed bits-to-remove data per channel per codec-block." << CRLF;
	ss_k = (Display_Width - (8 + 2) - (3 + 1)) / (3 * Global.channels);
	Titles[0] = "";
	Titles[1] = "";
	Titles[2] = "";
	Header = "";
	Make_Bars(Display_Width, 8, (3 * Global.channels - 1), ss_k, 3);
    ToOutput << top_bar << CRLF << "|  Time  |";

	for (rb_i = 0; rb_i < ss_k; rb_i ++)
		for (channel = 0; channel < Global.channels; channel ++)
		{
            ToOutput << '#' << channel;

			if (channel == Global.channels - 1)
			{
                ToOutput << '|';
			}
			else
			{
                ToOutput << ',';
			}
		}

    ToOutput << "Tot|" << CRLF << mid_bar << CRLF;
	ss_k = ss_k * Global.channels;
	btr_row_tot = 0;
	rb_j = 0;

	for (rb_i = 0; rb_i < Global.blocks_processed; rb_i ++)
		for (channel = 0; channel < Global.channels; channel ++)
		{
			rb_l = rb_j % ss_k;

			if (rb_l == 0)
			{
				time_string_make(time_string, double(rb_i) * Global.Codec_Block_size / Global.sample_rate);
				btr_row_tot = 0;
                ToOutput << '|' << time_string << '|';
			}

			rb_k = bit_removal_history[rb_j];
			btr_row_tot = btr_row_tot + rb_k;
            ToOutput << rb_k/*:2*/;

			if (channel == Global.channels - 1)
			{
                ToOutput << '|';
			}
			else
			{
                ToOutput << ',';
			}

			if (rb_l == ss_k - 1)
			{
                ToOutput << btr_row_tot/*:3:0*/ << '|' << CRLF;
			}

			++ rb_j;
		}

	rb_l = rb_j % ss_k;

	if ((rb_l < ss_k - 1) && (rb_l > 0))
	{
		for (rb_i = 0; rb_i <= (ss_k - rb_l - 1) / Global.channels; rb_i ++)
			for (channel = 0; channel < Global.channels; channel ++)
				if (channel < Global.channels - 1)
				{
                    ToOutput << "   ";
				}
				else
				{
                    ToOutput << "  |";
				}

        ToOutput << btr_row_tot/*:3:0*/ << '|' << CRLF;
	}

    ToOutput << mid_bar << CRLF;
}



#if defined(__GNUC__)
	static inline uint32_t clz(uint32_t x)
	{
        return __builtin_clz(x); // requires -march=486 on g++ command line on PC
	}
	static inline uint32_t ctz(uint32_t x)
	{
        return __builtin_ctz(x); // requires -march=486 on g++ command line on PC
	}
#elif defined(_MSC_VER)
	static inline uint32_t clz(uint32_t x)
	{
		unsigned long r = 0;
		_BitScanForward(&r, x);
		return r;
	}
	static inline uint32_t ctz(uint32_t x)
	{
		unsigned long r = 0;
		_BitScanReverse(&r, x);
		return r;
	}
#else
	static inline uint32_t popcnt(uint32_t x)
	{
		x -= ((x >> 1) & 0x55555555);
		x = (((x >> 2) & 0x33333333) + (x & 0x33333333));
		x = (((x >> 4) + x) & 0x0f0f0f0f);
		x += (x >> 8);
		x += (x >> 16);
		return x & 0x0000003f;
	}
	static inline uint32_t clz(uint32_t x)
	{
		x |= (x >> 1);
		x |= (x >> 2);
		x |= (x >> 4);
		x |= (x >> 8);
		x |= (x >> 16);
		return 32 - popcnt(x);
	}
	static inline uint32_t ctz(uint32_t x)
	{
		return popcnt((x & -int32_t(x)) - 1);
	}
#endif


inline int Bit_Scan_Left(int BSF_Int)
{
	// __asm {	bsf BSF_Int, Eax }
	return int(clz(uint32_t(BSF_Int)));
}

inline int Bit_Scan_Right(int BSR_Int)
{
	// __asm { bsr BSR_Int, Eax	}
	return int(ctz(uint32_t(BSR_Int)));
}


void lsb_analysis()
{
	int sa_i, sa_j;
	int temp_val;

	for (sa_j = 0; sa_j < Global.channels; sa_j ++)
	{
		DATA_block_value = 0;

		for (sa_i = 0; sa_i < BlockData.Size.This; sa_i ++)
		{
            temp_val = std::abs(BlockData.Pointer[2][sa_j][sa_i]);

			if (temp_val == 0)
			{
				++ DATA_sample_lsb[sa_j][0];
				++ DATA_sample_msb[sa_j][0];
			}
			else
			{
				++ DATA_sample_lsb[sa_j][Bit_Scan_Left(temp_val) + 1];
				++ DATA_sample_msb[sa_j][Bit_Scan_Right(temp_val) + 1];
			}

			DATA_block_value = DATA_block_value | temp_val;
		}

		if (DATA_block_value == 0)
		{
			++ DATA_block_lsb[sa_j][0];
			++ DATA_block_msb[sa_j][0];
		}
		else
		{
			++ DATA_block_lsb[sa_j][Bit_Scan_Left(DATA_block_value) + 1];
			++ DATA_block_msb[sa_j][Bit_Scan_Right(DATA_block_value) + 1];
		}

		BTRD_block_value = 0;

		for (sa_i = 0; sa_i < BlockData.Size.This; sa_i ++)
		{
			temp_val = std::abs(BlockData.BTRD[sa_j][sa_i]);

			if (temp_val == 0)
			{
				++ BTRD_sample_lsb[sa_j][0];
				++ BTRD_sample_msb[sa_j][0];
			}
			else
			{
				++ BTRD_sample_lsb[sa_j][Bit_Scan_Left(temp_val) + 1];
				++ BTRD_sample_msb[sa_j][Bit_Scan_Right(temp_val) + 1];
			}

			BTRD_block_value = BTRD_block_value | temp_val;
		}

		if (BTRD_block_value == 0)
		{
			++ BTRD_block_lsb[sa_j][0];
			++ BTRD_block_msb[sa_j][0];
		}
		else
		{
			++ BTRD_block_lsb[sa_j][Bit_Scan_Left(BTRD_block_value) + 1];
			++ BTRD_block_msb[sa_j][Bit_Scan_Right(BTRD_block_value) + 1];
		}

		CORR_block_value = 0;

		for (sa_i = 0; sa_i < BlockData.Size.This; sa_i ++)
		{
			temp_val = abs(BlockData.CORR[sa_j][sa_i]);

			if (temp_val == 0)
			{
				++ CORR_sample_lsb[sa_j][0];
				++ CORR_sample_msb[sa_j][0];
			}
			else
			{
				++ CORR_sample_lsb[sa_j][Bit_Scan_Left(temp_val) + 1];
				++ CORR_sample_msb[sa_j][Bit_Scan_Right(temp_val) + 1];
			}

			CORR_block_value = CORR_block_value | temp_val;
		}

		if (CORR_block_value == 0)
		{
			++ CORR_block_lsb[sa_j][0];
			++ CORR_block_msb[sa_j][0];
		}
		else
		{
			++ CORR_block_lsb[sa_j][Bit_Scan_Left(CORR_block_value) + 1];
			++ CORR_block_msb[sa_j][Bit_Scan_Right(CORR_block_value) + 1];
		}
	}
}


void Write_bitdist(std::ostream& ToOutput)
{
	int nd_i, nd_j;
	float this_bit_removed_percentage;
	int bits_removed_filled;

	if (! parameters.perchannel)
	{
		Header = "| Distribution of bits removed from each codec-block";
		Make_Bars(Display_Width, 3, 6, 1);
        ToOutput << CRLF << "Bits removed distribution." << CRLF << head_bar << CRLF << "|Bit" << Header << '|' << CRLF << mid_bar << CRLF;

		for (nd_i = 0; nd_i < Global.bits_per_sample; nd_i ++)
		{
			this_bit_removed_percentage = 0;

			for (nd_j = 0; nd_j < Global.channels; nd_j ++)
			{
				this_bit_removed_percentage += float(Stats.bits_removed[nd_j][nd_i] * Global.blocks_processed_recip * OneOver[Global.channels]);
			}

			bits_removed_filled = nround(this_bit_removed_percentage * bar_length);
            ToOutput << '|' << nd_i/*:3*/;
            ToOutput << '|' << (this_bit_removed_percentage * 100)/*:5:1*/ << "%|";
            ToOutput << std::string(bits_filled, bits_removed_filled)
                     << std::string(bits_empty, bar_length - bits_removed_filled)
                     << '|' << CRLF;
		}

        ToOutput << mid_bar << CRLF;
	}
	else
	{
		for (nd_i = 0; nd_i < Global.channels; nd_i ++)
		{
			Titles[nd_i] = std::string("| Ch. #") + char('0' + nd_i);
		}

		Make_Bars(Display_Width, 3, 6, Global.channels);
        ToOutput << CRLF << "Distribution of bits removed from each codec-block / channel." << CRLF << top_bar << CRLF << "|Bit";

		for (nd_i = 0; nd_i < Global.channels; nd_i ++)
		{
            ToOutput << Titles[nd_i];
		}

        ToOutput << '|' << CRLF << mid_bar << CRLF;

		for (nd_i = 0; nd_i < Global.bits_per_sample; nd_i ++)
		{
            ToOutput << '|' << nd_i/*:3*/ << '|';

			for (nd_j = 0; nd_j < Global.channels; nd_j ++)
			{
				this_bit_removed_percentage = float(Stats.bits_removed[nd_j][nd_i] * Global.blocks_processed_recip);
				bits_removed_filled = nround(this_bit_removed_percentage * bar_length);
                ToOutput << (this_bit_removed_percentage * 100)/*: 5: 1*/ << "%|";
                ToOutput << std::string(bits_filled, bits_removed_filled)
                         << std::string(bits_empty, bar_length - bits_removed_filled) << '|';
			}

            ToOutput << CRLF;
		}

        ToOutput << mid_bar << CRLF;
	}
}


void Write_blockdist(std::ostream& ToOutput)
{
	int nd_i, nd_j;
	int bits_filled_wave;
	int bits_filled_btrd;
	int bits_filled_corr;
	double this_DATA_percentage;
	double this_BTRD_percentage;
	double this_CORR_percentage;
	Titles[0] = "|Input lsb distribution";
	Titles[1] = "|Lossy lsb distribution";
	Titles[2] = "|LWCDF lsb distribution";
	Make_Bars(Display_Width, 3, 6, 3);
    ToOutput << CRLF << "Codec-block least significant bit (lsb) distribution." << CRLF << top_bar
             << CRLF << "|Bit" << Titles[0] << Titles[1] << Titles[2] << '|' << CRLF << mid_bar << CRLF;

	for (nd_i = 0; nd_i <= Global.bits_per_sample; nd_i ++)
	{
		this_DATA_percentage = 0;
		this_BTRD_percentage = 0;
		this_CORR_percentage = 0;

		for (nd_j = 0; nd_j < Global.channels; nd_j ++)
		{
			this_DATA_percentage += DATA_block_lsb[nd_j][nd_i] * Global.blocks_processed_recip;
			this_BTRD_percentage += BTRD_block_lsb[nd_j][nd_i] * Global.blocks_processed_recip;
			this_CORR_percentage += CORR_block_lsb[nd_j][nd_i] * Global.blocks_processed_recip;
		}

		bits_filled_wave = nround(this_DATA_percentage * bar_length);
		bits_filled_btrd = nround(this_BTRD_percentage * bar_length);
		bits_filled_corr = nround(this_CORR_percentage * bar_length);

		if (nd_i == 0)
		{
            ToOutput << "|NUL";
		}
		else
		{
            ToOutput << '|' << (nd_i - 1)/*: 3*/;
		}

        ToOutput << '|' << floattostrf(this_DATA_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_wave) << std::string(bits_empty, bar_length - bits_filled_wave) << '|';
        ToOutput << floattostrf(this_BTRD_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_btrd) << std::string(bits_empty, bar_length - bits_filled_btrd) << '|';
        ToOutput << floattostrf(this_CORR_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_corr) << std::string(bits_empty, bar_length - bits_filled_corr) << '|' << CRLF;
	}

	Titles[0] = "|Input msb distribution";
	Titles[1] = "|Lossy msb distribution";
	Titles[2] = "|LWCDF msb distribution";
	Make_Bars(Display_Width, 3, 6, 3);
    ToOutput << mid_bar << CRLF << CRLF << "Codec-block most significant bit (msb) distribution." << CRLF
             << top_bar << CRLF << "|Bit" << Titles[0] << Titles[1] << Titles[2] << '|' << CRLF << mid_bar << CRLF;

	for (nd_i = 0; nd_i <= Global.bits_per_sample; nd_i ++)
	{
		this_DATA_percentage = 0;
		this_BTRD_percentage = 0;
		this_CORR_percentage = 0;

		for (nd_j = 0; nd_j < Global.channels; nd_j ++)
		{
			this_DATA_percentage += DATA_block_msb[nd_j][nd_i] * Global.blocks_processed_recip;
			this_BTRD_percentage += BTRD_block_msb[nd_j][nd_i] * Global.blocks_processed_recip;
			this_CORR_percentage += CORR_block_msb[nd_j][nd_i] * Global.blocks_processed_recip;
		}

		bits_filled_wave = nround(this_DATA_percentage * bar_length);
		bits_filled_btrd = nround(this_BTRD_percentage * bar_length);
		bits_filled_corr = nround(this_CORR_percentage * bar_length);

		if (nd_i == 0)
		{
            ToOutput << "|NUL";
		}
		else
		{
            ToOutput << '|' << (nd_i - 1)/*: 3*/;
		}

        ToOutput << '|' << floattostrf(this_DATA_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_wave) << std::string(bits_empty, bar_length - bits_filled_wave) << '|';
        ToOutput << floattostrf(this_BTRD_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_btrd) << std::string(bits_empty, bar_length - bits_filled_btrd) << '|';
        ToOutput << floattostrf(this_CORR_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_corr) << std::string(bits_empty, bar_length - bits_filled_corr) << '|' << CRLF;
	}

    ToOutput << mid_bar << CRLF;
}


void Write_SampleDist(std::ostream& ToOutput)
{
	int nd_i, nd_j;
	int bits_filled_wave;
	int bits_filled_btrd;
	int bits_filled_corr;
	double this_DATA_percentage;
	double this_BTRD_percentage;
	double this_CORR_percentage;
	Titles[0] = "|Input lsb distribution";
	Titles[1] = "|Lossy lsb distribution";
	Titles[2] = "|LWCDF lsb distribution";
	Make_Bars(Display_Width, 3, 6, 3);
    ToOutput << CRLF << "Sample least significant bit (lsb) distribution." << CRLF << top_bar
             << CRLF << "|Bit" << Titles[0] << Titles[1] << Titles[2] << '|' << CRLF << mid_bar << CRLF;

	for (nd_i = 0; nd_i <= Global.bits_per_sample; nd_i ++)
	{
		this_DATA_percentage = 0;
		this_BTRD_percentage = 0;
		this_CORR_percentage = 0;

		for (nd_j = 0; nd_j < Global.channels; nd_j ++)
		{
			this_DATA_percentage += DATA_sample_lsb[nd_j][nd_i] * Global.samples_processed_recip;
			this_BTRD_percentage += BTRD_sample_lsb[nd_j][nd_i] * Global.samples_processed_recip;
			this_CORR_percentage += CORR_sample_lsb[nd_j][nd_i] * Global.samples_processed_recip;
		}

		bits_filled_wave = nround(this_DATA_percentage * bar_length);
		bits_filled_btrd = nround(this_BTRD_percentage * bar_length);
		bits_filled_corr = nround(this_CORR_percentage * bar_length);

		if (nd_i == 0)
		{
            ToOutput << "|NUL";
		}
		else
		{
            ToOutput << '|' << (nd_i - 1)/*: 3*/;
		}

        ToOutput << '|' << floattostrf(this_DATA_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_wave) << std::string(bits_empty, bar_length - bits_filled_wave) << '|';
        ToOutput << floattostrf(this_BTRD_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_btrd) << std::string(bits_empty, bar_length - bits_filled_btrd) << '|';
        ToOutput << floattostrf(this_CORR_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_corr) << std::string(bits_empty, bar_length - bits_filled_corr) << '|' << CRLF;
	}

	Titles[0] = "|Input msb distribution";
	Titles[1] = "|Lossy msb distribution";
	Titles[2] = "|LWCDF msb distribution";
	Make_Bars(Display_Width, 3, 6, 3);
    ToOutput << mid_bar << CRLF << CRLF << "Sample most significant bit (msb) distribution." << CRLF
             << top_bar << CRLF << "|Bit" << Titles[0] << Titles[1] << Titles[2] << '|' << CRLF
             << mid_bar << CRLF;

	for (nd_i = 0; nd_i <= Global.bits_per_sample; nd_i ++)
	{
		this_DATA_percentage = 0;
		this_BTRD_percentage = 0;
		this_CORR_percentage = 0;

		for (nd_j = 0; nd_j < Global.channels; nd_j ++)
		{
			this_DATA_percentage += DATA_sample_msb[nd_j][nd_i] * Global.samples_processed_recip;
			this_BTRD_percentage += BTRD_sample_msb[nd_j][nd_i] * Global.samples_processed_recip;
			this_CORR_percentage += CORR_sample_msb[nd_j][nd_i] * Global.samples_processed_recip;
		}

		bits_filled_wave = nround(this_DATA_percentage * bar_length);
		bits_filled_btrd = nround(this_BTRD_percentage * bar_length);
		bits_filled_corr = nround(this_CORR_percentage * bar_length);

		if (nd_i == 0)
		{
            ToOutput << "|NUL";
		}
		else
		{
            ToOutput << '|' << (nd_i - 1)/*: 3*/;
		}

        ToOutput << '|' << floattostrf(this_DATA_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_wave) << std::string(bits_empty, bar_length - bits_filled_wave) << '|';
        ToOutput << floattostrf(this_BTRD_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_btrd) << std::string(bits_empty, bar_length - bits_filled_btrd) << '|';
        ToOutput << floattostrf(this_CORR_percentage * 100, ffFixed, 99, 1)/*: 5*/ << "%|";
        ToOutput << std::string(bits_filled, bits_filled_corr) << std::string(bits_empty, bar_length - bits_filled_corr) << '|' << CRLF;
	}

    ToOutput << mid_bar << CRLF;
}


void Write_Histogram(std::ostream& ToOutput)
{
	int nd_i;
	int total_samples;
	int temp_val;

	int DATA_Max;
	int BTRD_Max;
	int CORR_Max;

	double DATA_val;
	double BTRD_val;
	double CORR_val;

	int DATA_len;
	int BTRD_len;
	int CORR_len;

	std::string DATA_bar;
	std::string BTRD_bar;
	std::string CORR_bar;

	Titles[0] = "|Input Value Histogram";
	Titles[1] = "|Lossy Value Histogram";
	Titles[2] = "|LWCDF Value Histogram";
	Make_Bars(Display_Width, 3, 7, 3);

    ToOutput << CRLF << "Sample Value Histogram." << CRLF << top_bar
             << CRLF << "|Bin" << Titles[0] << Titles[1] << Titles[2] << '|' << CRLF << mid_bar << CRLF;

	total_samples = 0;
	DATA_Max = 0;
	BTRD_Max = 0;
	CORR_Max = 0;

	for (nd_i = 0; nd_i <= history.Histogram_Length; nd_i ++)
	{
		temp_val = history.Histogram_DATA[nd_i];
		total_samples = total_samples + temp_val;

		if (temp_val > DATA_Max)
		{
			DATA_Max = temp_val;
		}

		temp_val = history.Histogram_BTRD[nd_i];

		if (temp_val > BTRD_Max)
		{
			BTRD_Max = temp_val;
		}

		temp_val = history.Histogram_CORR[nd_i];

		if (temp_val > CORR_Max)
		{
			CORR_Max = temp_val;
		}
	}

	for (nd_i = 0; nd_i <= history.Histogram_Length; nd_i ++)
	{
		temp_val = history.Histogram_DATA[nd_i];
		DATA_val = double(temp_val) / total_samples * 100;
		DATA_len = int(nlog2(temp_val) / nlog2(DATA_Max) * bar_length);
		DATA_bar = std::string(bits_filled, DATA_len) + std::string(bits_empty, bar_length - DATA_len);

		temp_val = history.Histogram_BTRD[nd_i];
		BTRD_val = double(temp_val) / total_samples * 100;
		BTRD_len = int(nlog2(temp_val) / nlog2(BTRD_Max) * bar_length);
		BTRD_bar = std::string(bits_filled, BTRD_len) + std::string(bits_empty, bar_length - BTRD_len);

		temp_val = history.Histogram_CORR[nd_i];
		CORR_val = double(temp_val) / total_samples * 100;
		CORR_len = int(nlog2(temp_val) / nlog2(CORR_Max) * bar_length);
		CORR_bar = std::string(bits_filled, CORR_len) + std::string(bits_empty, bar_length - CORR_len);
        ToOutput << '|' << (nd_i - history.Histogram_Offset)/*: 3*/
                 << '|' << DATA_val/*: 6: 2*/ << "%|" << DATA_bar
                 << '|' << BTRD_val/*: 6: 2*/ << "%|" << BTRD_bar
                 << '|' << CORR_val/*: 6: 2*/ << "%|" << CORR_bar
                 << '|' << CRLF;
	}

    ToOutput << mid_bar << CRLF;
}


void Write_FreqDist(std::ostream& ToOutput)
{
	int nt_i, nt_j;
	double nt_x, nt_m, nt_r;
	int fr_l, fr_h, fr_m;

	std::string inp_DATA_string;
	std::string Out_BTRD_String;
	std::string Out_CORR_String;

	double nt_inp_DATA_val;
	double nt_Out_BTRD_val;
	double nt_Out_CORR_val;

	int nt_inp_DATA_len;
	int nt_Out_BTRD_len;
	int nt_Out_CORR_len;

	char BinChar;

	if ((! parameters.postanalyse) && (Global.channels == 2))
	{
		nt_x = Global.blocks_processed_recip;
	}
	else
	{
		nt_x = OneOver[Global.channels] * Global.blocks_processed_recip;
	}

	nt_m = -db_lg2x20 * (Global.bits_per_sample + nlog2(1.5) * SINGLE_HALF);
	nt_r = 1.0 / nt_m;

	fr_l = std::max(1, nround(double(Global.lower_freq_limit) / Global.sample_rate * history.FFT.length));
	fr_h = nround(double(Global.upper_freq_limit) / Global.sample_rate * history.FFT.length);
	fr_m = nround(20000.0 / Global.sample_rate * history.FFT.length);

    ToOutput << CRLF << "Frequency Analysis of audio data." << CRLF;

	if (parameters.postanalyse)
	{
		Titles[0] = "| Input Average (dBFS)";
		Titles[1] = "| Lossy Average (dBFS)";
		Titles[2] = "| LWCDF Average (dBFS)";
		Make_Bars(Display_Width, 4, 6, 3, 6);

        ToOutput << top_bar << CRLF << "| Bin" << Titles[0] << Titles[1] << Titles[2] << "|Change|" << CRLF << mid_bar << CRLF;

		for (nt_i = 0; nt_i <= history.FFT.length_half; nt_i ++)
		{
			nt_inp_DATA_val = 0;
			nt_Out_BTRD_val = 0;
			nt_Out_CORR_val = 0;

			for (nt_j = 0; nt_j < Global.channels; nt_j ++)
			{
				nt_inp_DATA_val = nt_inp_DATA_val + history.DATA_results[nt_j][nt_i];
				nt_Out_BTRD_val = nt_Out_BTRD_val + history.BTRD_results[nt_j][nt_i];
				nt_Out_CORR_val = nt_Out_CORR_val + history.CORR_results[nt_j][nt_i];
			}

			nt_inp_DATA_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_inp_DATA_val * nt_x) * SINGLE_HALF - history.FFT.bit_length + 3 - Global.bits_per_sample));
			nt_Out_BTRD_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_Out_BTRD_val * nt_x) * SINGLE_HALF - history.FFT.bit_length + 3 - Global.bits_per_sample));
			nt_Out_CORR_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_Out_CORR_val * nt_x) * SINGLE_HALF - history.FFT.bit_length + 3 - Global.bits_per_sample));

			nt_inp_DATA_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_inp_DATA_val * nt_r)));
			nt_Out_BTRD_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_Out_BTRD_val * nt_r)));
			nt_Out_CORR_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_Out_CORR_val * nt_r)));

			inp_DATA_string = std::string(bits_filled, nt_inp_DATA_len) + std::string(bits_empty, bar_length - nt_inp_DATA_len);
			Out_BTRD_String = std::string(bits_filled, nt_Out_BTRD_len) + std::string(bits_empty, bar_length - nt_Out_BTRD_len);
			Out_CORR_String = std::string(bits_filled, nt_Out_CORR_len) + std::string(bits_empty, bar_length - nt_Out_CORR_len);

			if (nt_i == fr_l)
			{
				BinChar = 'L';
			}
			else if (nt_i == fr_h)
			{
				BinChar = 'U';
			}
			else if (nt_i == fr_m)
			{
				BinChar = '*';
			}
			else
			{
				BinChar = ' ';
			}

			if (nt_i == 0)
			{
                ToOutput << "| DC ";
			}
			else if (nt_i == PowersOf.TwoInt[history.FFT.bit_length - 1])
			{
                ToOutput << "|FS/2";
			}
			else
			{
                ToOutput << '|' << BinChar << nt_i/*: 3*/;
			}

            ToOutput << '|' << nt_inp_DATA_val/*: 6: 1*/ << '|' << inp_DATA_string << '|' << nt_Out_BTRD_val/*: 6: 1*/ << '|' << Out_BTRD_String << '|' << nt_Out_CORR_val/*: 6: 1*/ << '|' << Out_CORR_String << '|' << BinChar << nround10(nt_Out_BTRD_val - nt_inp_DATA_val, 2)/*: 5: 2*/ << '|' << CRLF;
		}

        ToOutput << mid_bar << CRLF;
	}
	else if (Global.channels != 2)
	{
		Titles[0] = "| Input Average (dBFS)";
		Make_Bars(Display_Width, 4, 6, 1);
        ToOutput << top_bar << CRLF << "| Bin" << Titles[0] << std::string("|") << CRLF << mid_bar << CRLF;

		for (nt_i = 0; nt_i <= history.FFT.length_half; nt_i ++)
		{
			nt_inp_DATA_val = 0;

			for (nt_j = 0; nt_j < Global.channels; nt_j ++)
			{
				nt_inp_DATA_val = nt_inp_DATA_val + history.DATA_results[nt_j][nt_i];
			}

			nt_inp_DATA_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_inp_DATA_val * nt_x) * SINGLE_HALF - history.FFT.bit_length + 3 - Global.bits_per_sample));
			nt_inp_DATA_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_inp_DATA_val * nt_r)));
			inp_DATA_string = std::string(bits_filled, nt_inp_DATA_len) + std::string(bits_empty, bar_length - nt_inp_DATA_len);

			if (nt_i == fr_l)
			{
				BinChar = 'L';
			}
			else if (nt_i == fr_h)
			{
				BinChar = 'U';
			}
			else if (nt_i == fr_m)
			{
				BinChar = '*';
			}
			else
			{
				BinChar = ' ';
			}

			if (nt_i == 0)
			{
                ToOutput << "| DC ";
			}
			else if (nt_i == PowersOf.TwoInt[history.FFT.bit_length - 1])
			{
                ToOutput << "|FS/2";
			}
			else
			{
                ToOutput << '|' << BinChar << nt_i/*: 3*/;
			}

            ToOutput << '|' << nt_inp_DATA_val/*: 6: 1*/ << '|' << inp_DATA_string << '|' << CRLF;
		}

        ToOutput << mid_bar << CRLF;
	}
	else
	{
		Titles[0] = "| Input Average (dBFS) Channel #0";
		Titles[1] = "| Input Average (dBFS) Channel #1";
		Make_Bars(Display_Width, 4, 6, 2);
        ToOutput << top_bar << CRLF << "| Bin" << Titles[0] << Titles[1] << std::string("|") << CRLF << mid_bar << CRLF;

		for (nt_i = 0; nt_i <= history.FFT.length_half; nt_i ++)
		{
			nt_inp_DATA_val = history.DATA_results[0][nt_i];
			nt_Out_BTRD_val = history.DATA_results[1][nt_i];
			nt_inp_DATA_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_inp_DATA_val * nt_x) * SINGLE_HALF - history.FFT.bit_length + 3 - Global.bits_per_sample));
			nt_Out_BTRD_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_Out_BTRD_val * nt_x) * SINGLE_HALF - history.FFT.bit_length + 3 - Global.bits_per_sample));
			nt_inp_DATA_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_inp_DATA_val * nt_r)));
			nt_Out_BTRD_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_Out_BTRD_val * nt_r)));
			inp_DATA_string = std::string(bits_filled, nt_inp_DATA_len) + std::string(bits_empty, bar_length - nt_inp_DATA_len);
			Out_BTRD_String = std::string(bits_filled, nt_Out_BTRD_len) + std::string(bits_empty, bar_length - nt_Out_BTRD_len);

			if (nt_i == fr_l)
			{
				BinChar = 'L';
			}
			else if (nt_i == fr_h)
			{
				BinChar = 'U';
			}
			else if (nt_i == fr_m)
			{
				BinChar = '*';
			}
			else
			{
				BinChar = ' ';
			}

			if (nt_i == 0)
			{
                ToOutput << "| DC ";
			}
			else if (nt_i == PowersOf.TwoInt[history.FFT.bit_length - 1])
			{
                ToOutput << "|FS/2";
			}
			else
			{
                ToOutput << '|' << BinChar << nt_i/*: 3*/;
			}

            ToOutput << '|' << nt_inp_DATA_val/*: 6: 1*/ << '|' << inp_DATA_string << '|' << nt_Out_BTRD_val/*: 6: 1*/ << '|' << Out_BTRD_String << '|' << CRLF;
		}

        ToOutput << mid_bar << CRLF;
	}

    ToOutput << "Legend: L = Lower Frequency Calculation Limit ( " << floattostrf(Global.lower_freq_limit * OneOver[1000], ffFixed, 99, 3) << "kHz);" << CRLF
             << "        U = Upper Frequency Calculation Limit ( " << floattostrf(Global.upper_freq_limit * OneOver[1000], ffFixed, 99, 3) << "kHz);" << CRLF
             << "        * = 20.000kHz" << CRLF;
}


void Write_LongDist(std::ostream& ToOutput)
{
	int nt_i, nt_j;
	double nt_x, nt_m, nt_r;
	int fr_l, fr_h, fr_m;
	std::string inp_DATA_string;
	std::string Out_BTRD_String;
	std::string Out_CORR_String;
	double nt_inp_DATA_val;
	double nt_Out_BTRD_val;
	double nt_Out_CORR_val;
	int nt_inp_DATA_len;
	int nt_Out_BTRD_len;
	int nt_Out_CORR_len;
	char BinChar;
	nt_x = OneOver[Global.channels] * Global.blocks_processed_recip;
	nt_m = -db_lg2x20 * (Global.bits_per_sample + nlog2(1.5) * SINGLE_HALF);
	nt_r = 1.0 / nt_m;
	fr_l = std::max(1, nround(double(Global.lower_freq_limit) / Global.sample_rate * LongDist.FFT.length));
	fr_h = nround(double(Global.upper_freq_limit) / Global.sample_rate * LongDist.FFT.length);
	fr_m = nround(20000.0 / Global.sample_rate * LongDist.FFT.length);
    ToOutput << CRLF << "Frequency Analysis of audio data." << CRLF;
	Titles[0] = "| Input Average (dBFS)";
	Titles[1] = "| Lossy Average (dBFS)";
	Titles[2] = "| LWCDF Average (dBFS)";
	Make_Bars(Display_Width, 4, 6, 3, 6);
    ToOutput << top_bar << CRLF << "| Bin" << Titles[0] << Titles[1] << Titles[2] << "|Change|" << CRLF << mid_bar << CRLF;

	for (nt_i = 0; nt_i <= LongDist.FFT.length_half; nt_i ++)
	{
		nt_inp_DATA_val = 0;
		nt_Out_BTRD_val = 0;
		nt_Out_CORR_val = 0;

		for (nt_j = 0; nt_j < Global.channels; nt_j ++)
		{
			nt_inp_DATA_val = nt_inp_DATA_val + LongDist.DATA_results[nt_j][nt_i];
			nt_Out_BTRD_val = nt_Out_BTRD_val + LongDist.BTRD_results[nt_j][nt_i];
			nt_Out_CORR_val = nt_Out_CORR_val + LongDist.CORR_results[nt_j][nt_i];
		}

		nt_inp_DATA_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_inp_DATA_val * nt_x) * SINGLE_HALF - LongDist.FFT.bit_length + 3 - Global.bits_per_sample));
		nt_Out_BTRD_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_Out_BTRD_val * nt_x) * SINGLE_HALF - LongDist.FFT.bit_length + 3 - Global.bits_per_sample));
		nt_Out_CORR_val = std::max(nt_m, db_lg2x20 * (nlog2(nt_Out_CORR_val * nt_x) * SINGLE_HALF - LongDist.FFT.bit_length + 3 - Global.bits_per_sample));
		nt_inp_DATA_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_inp_DATA_val * nt_r)));
		nt_Out_BTRD_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_Out_BTRD_val * nt_r)));
		nt_Out_CORR_len = bar_length - std::min(bar_length, std::max(0, nround(bar_length * nt_Out_CORR_val * nt_r)));
		inp_DATA_string = std::string(bits_filled, nt_inp_DATA_len) + std::string(bits_empty, bar_length - nt_inp_DATA_len);
		Out_BTRD_String = std::string(bits_filled, nt_Out_BTRD_len) + std::string(bits_empty, bar_length - nt_Out_BTRD_len);
		Out_CORR_String = std::string(bits_filled, nt_Out_CORR_len) + std::string(bits_empty, bar_length - nt_Out_CORR_len);

		if (nt_i == fr_l)
		{
			BinChar = 'L';
		}
		else if (nt_i == fr_h)
		{
			BinChar = 'U';
		}
		else if (nt_i == fr_m)
		{
			BinChar = '*';
		}
		else
		{
			BinChar = ' ';
		}

		if (nt_i == 0)
		{
            ToOutput << "| DC ";
		}
		else if (nt_i == PowersOf.TwoInt[LongDist.FFT.bit_length - 1])
		{
            ToOutput << "|FS/2";
		}
		else
		{
            ToOutput << '|' << BinChar << nt_i/*: 3*/;
		}

        ToOutput << '|' << nt_inp_DATA_val/*: 6: 1*/ << '|' << inp_DATA_string
                 << '|' << nt_Out_BTRD_val/*: 6: 1*/ << '|' << Out_BTRD_String
                 << '|' << nt_Out_CORR_val/*: 6: 1*/ << '|' << Out_CORR_String
                 << '|' << BinChar << nround10(nt_Out_BTRD_val - nt_inp_DATA_val, 2)/*: 5: 2*/ << '|' << CRLF;
	}

    ToOutput << mid_bar << CRLF;
    ToOutput << "Legend: L = Lower Frequency Calculation Limit (" << floattostrf(Global.lower_freq_limit * OneOver[1000], ffFixed, 99, 3) << "kHz);" << CRLF
             << "        U = Upper Frequency Calculation Limit (" << floattostrf(Global.upper_freq_limit * OneOver[1000], ffFixed, 99, 3) << "kHz);" << CRLF
             << "        * = 20.000kHz" << CRLF;
}


void Write_Min_Bin_Dist(std::ostream& ToOutput)
{
	int nt_i, nt_j;
	int Bar_Full_Len;
	double Old_Percent;
	double New_Percent;
	double Alt_Percent;
	double Tot_Old_Percent;
	double Tot_New_Percent;
	double Tot_Alt_Percent;
	std::string Old_Bar_Str;
	std::string New_Bar_Str;
	std::string Alt_Bar_Str;
	int Local_Max;
	double Percent_Used [ 9 ];
	int fr_l, fr_h, fr_m;
	char BinChar;

	if (parameters.spread == 2)
	{
		Titles[0] = "| Old Minimum Value Used";
		Titles[1] = "| New Minimum Value Used";
		Titles[2] = "| Minimum Value Used";
		Make_Bars(Display_Width, 4, 7, 3);

        // Empty Loop. // TY
/*
		for (nt_j = 1; nt_j <= PRECALC_ANALYSES; nt_j ++)
		{
			if ((settings.FFT_analysis_switches & Global.analysis.length[nt_j]) > 0) ;
		}
*/
		for (nt_j = 1; nt_j <= PRECALC_ANALYSES; nt_j ++)
			if ((settings.FFT_analysis_switches & Global.analysis.length[nt_j]) > 0)
			{
				Local_Max = 0;
				Tot_Old_Percent = 0;
				Tot_New_Percent = 0;
				Tot_Alt_Percent = 0;

				for (nt_i = 0; nt_i <= spreading.Bins.hi[Global.analysis.bits[nt_j]] + 1; nt_i ++)
				{
					Local_Max = std::max(Local_Max, process.Old_Min_Used_History[nt_j][nt_i] + process.New_Min_Used_History[nt_j][nt_i]);
				}

				fr_l = spreading.Bins.Lo[Global.analysis.bits[nt_j]];
				fr_h = spreading.Bins.hi[Global.analysis.bits[nt_j]];
				fr_m = nround(20000.0 / Global.sample_rate * Global.analysis.length[nt_j]);
                ToOutput << CRLF << "Spreading Algorithm Results : Minima, FFT Length = " << IntToStr(Global.analysis.length[nt_j])
                         << CRLF << top_bar << CRLF << "|Bin " << Titles[0] << Titles[1] << std::string("|") << CRLF << mid_bar << CRLF;

				for (nt_i = 0; nt_i <= spreading.Bins.hi[Global.analysis.bits[nt_j]] + 1; nt_i ++)
				{
					if (nt_i == fr_l)
					{
						BinChar = 'L';
					}
					else if (nt_i == fr_h)
					{
						BinChar = 'U';
					}
					else if (nt_i == fr_m)
					{
						BinChar = '*';
					}
					else
					{
						BinChar = ' ';
					}

					if (nt_i == 0)
					{
                        ToOutput << "| DC ";
					}
					else if (nt_i == Global.analysis.length[nt_j])
					{
                        ToOutput << "|FS/2";
					}
					else
					{
                        ToOutput << '|' << BinChar << nt_i/*: 3*/;
					}

					Old_Percent = double(process.Old_Min_Used_History[nt_j][nt_i]) / process.Analyses_Completed[nt_j] * 100;
					New_Percent = double(process.New_Min_Used_History[nt_j][nt_i]) / process.Analyses_Completed[nt_j] * 100;
					Alt_Percent = Old_Percent + New_Percent;
					Bar_Full_Len = nround(double(process.Old_Min_Used_History[nt_j][nt_i]) / Local_Max * bar_length);
					Old_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
					Bar_Full_Len = nround(double(process.New_Min_Used_History[nt_j][nt_i]) / Local_Max * bar_length);
					New_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
					Bar_Full_Len = nround(double(process.Old_Min_Used_History[nt_j][nt_i] + process.New_Min_Used_History[nt_j][nt_i]) / Local_Max * bar_length);
					Alt_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
					Tot_Old_Percent = Tot_Old_Percent + Old_Percent;
					Tot_New_Percent = Tot_New_Percent + New_Percent;
					Tot_Alt_Percent = Tot_Alt_Percent + Alt_Percent;
                    ToOutput << '|' << Old_Percent/*: 6: 2*/ << "%|" << Old_Bar_Str
                             << '|' << New_Percent/*: 6: 2*/ << "%|" << New_Bar_Str
                             << '|' << Alt_Percent/*: 6: 2*/ << "%|" << Alt_Bar_Str << '|' << CRLF;
				}

                ToOutput << mid_bar << CRLF;
				Bar_Full_Len = nround(Tot_Old_Percent * bar_length * OneOver[100]);
				Old_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
				Bar_Full_Len = nround(Tot_New_Percent * bar_length * OneOver[100]);
				New_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
				Bar_Full_Len = nround(Tot_Alt_Percent * bar_length * OneOver[100]);
				Alt_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
                ToOutput << "|Tot.|" << Tot_Old_Percent/*: 6: 2*/ << "%|" << Old_Bar_Str
                         <<      '|' << Tot_New_Percent/*: 6: 2*/ << "%|" << New_Bar_Str
                         <<      '|' << Tot_Alt_Percent/*: 6: 2*/ << "%|" << Alt_Bar_Str << '|' << CRLF;
                ToOutput << mid_bar << CRLF;
			}
	}

	Titles[0] = "| Minimum used per codec-block";
	Make_Bars(Display_Width, 4, 7, 1);
    ToOutput << CRLF << "Minimum used to determine bits-to-remove, by FFT length:" << CRLF << top_bar
             << CRLF << "|FFT " << Titles[0] << std::string("|") << CRLF << mid_bar << CRLF;

	for (nt_j = 1; nt_j <= PRECALC_ANALYSES; nt_j ++)
		if ((settings.FFT_analysis_switches & Global.analysis.length[nt_j]) > 0)
		{
			Percent_Used[nt_j] = 0;

			for (nt_i = 0; nt_i < Global.channels; nt_i ++)
			{
				Percent_Used[nt_j] = Percent_Used[nt_j] + results.minima[nt_i][nt_j];
			}

			Percent_Used[nt_j] = Percent_Used[nt_j] * Global.blocks_processed_recip * OneOver[Global.channels] * 100;
			Bar_Full_Len = nround(Percent_Used[nt_j] * bar_length * OneOver[100]);
			Old_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
            ToOutput << '|' << Global.analysis.length[nt_j]/*: 4*/
                     << '|' << Percent_Used[nt_j]/*: 6: 2*/
                     << "%|" << Old_Bar_Str << '|' << CRLF;
		}

	nt_j = 7;
	Percent_Used[nt_j] = 0;

	for (nt_i = 0; nt_i < Global.channels; nt_i ++)
	{
		Percent_Used[nt_j] += results.minima[nt_i][nt_j];
	}

	Percent_Used[nt_j] = Percent_Used[nt_j] * Global.blocks_processed_recip * OneOver[Global.channels] * 100;
	Bar_Full_Len = nround(Percent_Used[nt_j] * bar_length * OneOver[100]);
	Old_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
    ToOutput << "|Stat|" << Percent_Used[nt_j]/*: 6: 2*/ << "%|" << Old_Bar_Str << '|' << CRLF;
	nt_j = 8;
	Percent_Used[nt_j] = 0;

	for (nt_i = 0; nt_i < Global.channels; nt_i ++)
	{
		Percent_Used[nt_j] = Percent_Used[nt_j] + results.minima[nt_i][nt_j];
	}

	Percent_Used[nt_j] = Percent_Used[nt_j] * Global.blocks_processed_recip * OneOver[Global.channels] * 100;
	Bar_Full_Len = nround(Percent_Used[nt_j] * bar_length * OneOver[100]);
	Old_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
    ToOutput << "|Dyn.|" << Percent_Used[nt_j]/*: 6: 2*/ << "%|" << Old_Bar_Str << '|' << CRLF;
    ToOutput << mid_bar << CRLF;
	Titles[0] = "| Old Minimum Value Used";
	Titles[1] = "| New Minimum Value Used";
	Titles[2] = "| Average Value Used";
	Make_Bars(Display_Width, 4, 7, 3);
    ToOutput << CRLF << "Spreading Algorithm Results : Minima, per analysis" << CRLF << top_bar
             << CRLF << "|FFT " << Titles[0] << Titles[1] << Titles[2] << "|" << CRLF << mid_bar << CRLF;

	for (nt_j = 1; nt_j <= PRECALC_ANALYSES; nt_j ++)
		if ((settings.FFT_analysis_switches & Global.analysis.length[nt_j]) > 0)
		{
			Old_Percent = double(process.Old_Min_Used[nt_j]) / process.Analyses_Completed[nt_j] * 100;
			New_Percent = double(process.New_Min_Used[nt_j]) / process.Analyses_Completed[nt_j] * 100;
			Alt_Percent = double(process.Alt_Ave_Used[nt_j]) / process.Analyses_Completed[nt_j] * 100;
			Bar_Full_Len = nround(Old_Percent * bar_length * OneOver[100]);
			Old_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
			Bar_Full_Len = nround(New_Percent * bar_length * OneOver[100]);
			New_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
			Bar_Full_Len = nround(Alt_Percent * bar_length * OneOver[100]);
			Alt_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
            ToOutput << '|' << Global.analysis.length[nt_j]/*: 4*/
                     << '|' << Old_Percent/*: 6: 2*/ << "%|" << Old_Bar_Str
                     << '|' << New_Percent/*: 6: 2*/ << "%|" << New_Bar_Str
                     << '|' << Alt_Percent/*: 6: 2*/ << "%|" << Alt_Bar_Str
                     << '|' << CRLF;
		}

    ToOutput << mid_bar << CRLF;
	Titles[0] = "| BTR > Static-Max-Bits-To-Remove";
	Titles[1] = "| BTR > Dynamic-Max-Bits-To-Remove";
	Make_Bars(Display_Width, 4, 7, 2);
    ToOutput << CRLF << "Minimum_Bits_To_Keep Results (BTR limited to relevant maximum)" << CRLF << top_bar
             << CRLF << "|FFT " << Titles[0] << Titles[1] << '|' << CRLF << mid_bar << CRLF;

	for (nt_j = 1; nt_j <= PRECALC_ANALYSES; nt_j ++)
		if ((settings.FFT_analysis_switches & Global.analysis.length[nt_j]) > 0)
		{
			Old_Percent = double(process.Over_Static[nt_j]) / process.Analyses_Completed[nt_j] * 100;
			New_Percent = double(process.Over_Dynamic[nt_j]) / process.Analyses_Completed[nt_j] * 100;
			Bar_Full_Len = nround(Old_Percent * bar_length * OneOver[100]);
			Old_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
			Bar_Full_Len = nround(New_Percent * bar_length * OneOver[100]);
			New_Bar_Str = std::string(bits_filled, Bar_Full_Len) + std::string(bits_empty, bar_length - Bar_Full_Len);
            ToOutput << '|' << Global.analysis.length[nt_j]/*: 4*/
                     << '|' << Old_Percent/*: 6: 2*/ << "%|" << Old_Bar_Str
                     << '|' << New_Percent/*: 6: 2*/ << "%|" << New_Bar_Str
                     << '|' << CRLF;
		}

    ToOutput << mid_bar << CRLF;
}


void Process_Output()
{
	if ((parameters.verbosity) && (! parameters.silent) && ((Global.last_print == 0) || (Global.blocks_processed % Global.output_blocks == 0)))
	{
		gettimer();
		time_string_make(strings.Elapsed, timer.Elapsed);
        ErrOutput << '\x0d' << "Progress  :";

		if (Global.wave_size == MAX_WAVE_SIZE)
		{
			size_string_make(strings.Size, 1.0 * Global.samples_processed * Global.channels * Global.bytes_per_sample);
            ErrOutput << ' ' << strings.Size << ';';
		}
		else
		{
            ErrOutput << (double(Global.samples_processed) / Global.total_samples * 100)/*: 6: 2*/ << "%;";
		}

        ErrOutput << (OneOver[Global.channels] * Global.bits_removed * Global.blocks_processed_recip)/*: 7: 4*/ << " bits;"
                  << Global.processing_rate/*: 6: 2*/ << "x;";
        ErrOutput << ' ' << strings.Elapsed;

		if (Global.samples_processed > 0)
		{
			time_string_make(strings.estimate, std::min(86399.0, double(Global.total_samples) / Global.samples_processed * timer.Elapsed));
            ErrOutput << '/' << strings.estimate << ' ';
		}

		Global.last_print = Global.blocks_processed;
	}
}


void Write_Results(std::ostream& ToOutput)
{
	double bits_lost;
    ToOutput << "Results   :" << (OneOver[Global.channels] * Global.bits_removed * Global.blocks_processed_recip)/*: 7: 4*/ << " bits; "
             << Global.processing_rate/*: 0: 2*/ << "x; " << strings.Elapsed;

	if (FFTW_Initialised())
	{
        ToOutput << "; [F]";
	}
	else
	{
        ToOutput << "; [I]";
	}

	if (Global.jump_location != 0)
	{
        ToOutput << "; J:" << CardinalToHex(Global.jump_location) << ';' << (Global.jump_location & 31);
	}

    ToOutput << CRLF;

	if (Global.bits_lost + Global.clipped_samples > 0)
	{
        ToOutput << "Clipping  :";
		bits_lost = nround10(Global.bits_lost * OneOver[Global.channels] * Global.blocks_processed_recip, 4);

		if (bits_lost > 0.5)
		{
            ToOutput << bits_lost/*: 7: 4*/ << " bits lost,";
		}

        ToOutput << ' ' << Global.clipped_samples << " clip";

		if (Global.clipped_samples != 1)
		{
            ToOutput << 's';
		}

		if (bits_lost > 0)
		{
            ToOutput << " remaining." << CRLF;
		}
		else
		{
            ToOutput << '.' << CRLF;
		}
	}

	if ((parameters.adaptive_shaping != -1) && (Skipped_Filters() > 0))
	{
        ToOutput << "Filters   : " << Skipped_Filters() << " adaptive noise shaping filter";

		if (Skipped_Filters() != 1)
		{
            ToOutput << 's';
		}

        ToOutput << " skipped." << CRLF;
	}

	Global.samples_processed_recip = OneOver[Global.channels] / Global.samples_processed;

	if (parameters.bitdist)
	{
		Write_bitdist(ToOutput);
	}

	if (parameters.blockdist)
	{
		Write_blockdist(ToOutput);
	}

	if (parameters.sampledist)
	{
		Write_SampleDist(ToOutput);
	}

	if (parameters.freqdist)
	{
		Write_FreqDist(ToOutput);
	}

	if (parameters.longdist)
	{
		Write_LongDist(ToOutput);
	}

	if (parameters.detail)
	{
		remove_bits_detailed_output(ToOutput);
	}

	if (parameters.spread != -1)
	{
		Write_Min_Bin_Dist(ToOutput);
	}

	if (parameters.histogram)
	{
		Write_Histogram(ToOutput);
	}

	flush(ToOutput);
}


void write_cleanup()
{
	gettimer();
	time_string_make(strings.Elapsed, timer.Elapsed);

	if (! parameters.silent)
	{
		if (parameters.verbosity)
		{
            ErrOutput << "\x0d                                                                               \x0d";
            Write_Results(ErrOutput);
		}
		else
		{
            ErrOutput << (OneOver[Global.channels] * Global.bits_removed * Global.blocks_processed_recip)/*: 7: 4*/ << ';'
                      << Global.processing_rate/*: 6: 2*/ << "x; " << strings.Elapsed;

			if (FFTW_Initialised())
			{
                ErrOutput << "; [F]";
			}
			else
			{
                ErrOutput << "; [I]";
			}

			if ((parameters.adaptive_shaping != -1) && (Skipped_Filters() > 0))
			{
                ErrOutput << "; F" << Skipped_Filters();
			}

			if (Global.jump_location != 0)
			{
                ErrOutput << "; J" << (Global.jump_location & 15);
			}

            ErrOutput << CRLF;
		}
	}

	if (parameters.writetolog)
	{
		time_string_make(strings.time, double(Global.samples_processed) / Global.sample_rate);
		size_string_make(strings.Size, 1.0 * Global.samples_processed * Global.channels * Global.bytes_per_sample);
		open_log_file();
        LogOutput << version_string << strings.version_short << lossyWAVHead1
                  << "Processed : " << strings.datestamp << CRLF
                  << "Settings  : " << strings.parameter << CRLF
                  << "Filename  : " << WAVFilePrintName() << CRLF
                  << "File Info : " << (Global.sample_rate / 1000.0)/*: 0: 2*/ << "kHz; "
                  << Global.channels << " channel; " << Global.bits_per_sample << " bit, "
                  << strings.time << ", " << strings.Size << CRLF;
		Write_Results(LogOutput);
        LogOutput << CRLF;
		close_log_file();
	}
}


void nOutput_Init()
{
	int nd_i, nd_j;

	for (nd_j = 0; nd_j < MAX_CHANNELS; nd_j ++)
		for (nd_i = 0; nd_i <= 33; nd_i ++)
		{
			DATA_block_lsb[nd_j][nd_i] = 0;
			DATA_block_msb[nd_j][nd_i] = 0;
			DATA_sample_lsb[nd_j][nd_i] = 0;
			DATA_sample_msb[nd_j][nd_i] = 0;
			BTRD_block_lsb[nd_j][nd_i] = 0;
			BTRD_block_msb[nd_j][nd_i] = 0;
			BTRD_sample_lsb[nd_j][nd_i] = 0;
			BTRD_sample_msb[nd_j][nd_i] = 0;
			CORR_block_lsb[nd_j][nd_i] = 0;
			CORR_block_msb[nd_j][nd_i] = 0;
			CORR_sample_lsb[nd_j][nd_i] = 0;
			CORR_sample_msb[nd_j][nd_i] = 0;
		}
}

// initialization: todo
static struct Init
{
	Init() {
		// initialization: todo
		nOutput_Init();

		if (parameters.detail)
		{
			bit_removal_history = new unsigned char[ bit_removal_history_array_size ];
		}

		if (parameters.width == -1)
		{
			Display_Width = 79;
		}
		else
		{
			Display_Width = parameters.width;
		}
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
		if (bit_removal_history != NULL)
		{
			delete[] bit_removal_history;
		}
	}
} init;


