//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include "nRemoveBits.h"
#include "nSGNS.h"

static const char UnitName [] = "nRemoveBits";

// Globals
remove_bits_data_type remove_bits_data;
int sample_maxima [ BITS_TO_CALCULATE ];


struct RemoveData_type
{
	double this_two_power;
	double this_two_power_recip;
	int this_min_sample;
	int this_max_sample;
	int64_t this_temp_sample;
}
RemoveData;


int Limited_Sample(double input_sample)
{
	int result;

	if (input_sample > RemoveData.this_max_sample)
	{
		input_sample = RemoveData.this_max_sample;
		++ remove_bits_data.clips;
	}
	else if (input_sample < RemoveData.this_min_sample)
	{
		input_sample = RemoveData.this_min_sample;
		++ remove_bits_data.clips;
	}

	result = int(input_sample);
	return result;
}


void Remove_Bits_Proc_Adaptive_Noise_Shaping_On()
{
	int rb_i;
	int64_t signal;
	double scaled;
	int64_t output;
	remove_bits_data.clips = 0;
	RemoveData.this_max_sample = sample_maxima[remove_bits_data.bits_removed];
	RemoveData.this_two_power = double(PowersOf.TwoInt[remove_bits_data.bits_removed]);
	RemoveData.this_two_power_recip = PowersOf.Two[- remove_bits_data.bits_removed];
	Restore_Filter();

	for (rb_i = 0; rb_i < BlockData.Size.This; ++rb_i)
	{
		signal = BlockData.Pointer[2][Global.channel][rb_i];
		scaled = signal * settings.scaling_factor;
		output = int64_t(Limited_Sample(nround((scaled + Warped_Lattice_Filter_Evaluate()) * RemoveData.this_two_power_recip) * RemoveData.this_two_power));
		Warped_Lattice_Filter_Update(output - scaled);
		BlockData.BTRD[Global.channel][rb_i] = int(output);
		BlockData.CORR[Global.channel][rb_i] = ((int) signal - nround(output * settings.scaling_factor_inv));
	}
}


void Remove_Bits_Proc_Shaping_Off()
{
	int rb_i;
	int64_t signal;
	double scaled;
	int64_t output;
	remove_bits_data.clips = 0;
	RemoveData.this_max_sample = sample_maxima[remove_bits_data.bits_removed];
	RemoveData.this_two_power = double(PowersOf.TwoInt[remove_bits_data.bits_removed]);
	RemoveData.this_two_power_recip = PowersOf.Two[- remove_bits_data.bits_removed];

	for (rb_i = 0; rb_i < BlockData.Size.This; ++rb_i)
	{
		signal = BlockData.Pointer[2][Global.channel][rb_i];
		scaled = signal * settings.scaling_factor;
		output = int64_t(Limited_Sample(nround(scaled * RemoveData.this_two_power_recip) * RemoveData.this_two_power));
		BlockData.BTRD[Global.channel][rb_i] = int(output);
		BlockData.CORR[Global.channel][rb_i] = int(signal - nround(output * settings.scaling_factor_inv));
	}
}


void Remove_Bits()
{
	remove_bits_data.bits_removed = remove_bits_data.bits_to_remove;

	do
	{
		if ((parameters.shaping != -1) || ((parameters.adaptive_shaping != -1) && (Filter_Valid())))
		{
			Remove_Bits_Proc_Adaptive_Noise_Shaping_On();
		}
		else
		{
			Remove_Bits_Proc_Shaping_Off();
		}

		if (remove_bits_data.clips <= settings.maximum_clips_per_channel)
		{
			return; // break; ? // TY
		}
		else
		{
			remove_bits_data.bits_removed = remove_bits_data.bits_removed - 1;
		}
	}
	while (!((remove_bits_data.bits_removed < 0) || (remove_bits_data.clips <= settings.maximum_clips_per_channel)));

	if (remove_bits_data.bits_removed < 0)
	{
		remove_bits_data.bits_removed = 0;
	}
}


void nRemoveBits_Init()
{
	int fbi_i;
	RemoveData.this_max_sample = PowersOf.TwoM1[Global.bits_per_sample - 1];
	RemoveData.this_min_sample = -PowersOf.TwoInt[Global.bits_per_sample - 1];

	for (fbi_i = 0; fbi_i < BITS_TO_CALCULATE; ++fbi_i)
	{
		sample_maxima[fbi_i] = long(RemoveData.this_max_sample ^ PowersOf.TwoM1[fbi_i]); // ^ = bitwise xor // TY
	}

	settings.static_maximum_bits_to_remove = std::max(0, Global.bits_per_sample - settings.static_minimum_bits_to_keep);
}
// initialization: todo
// finalization: todo Write ( Erroutput, UnitName, CRLF );
