//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include "nShiftBlocks.h"
#include "nCore.h"
#include "nMaths.h"

static const char UnitName [] = "nShiftBlocks";

static double Channel_RMS(int this_channel)
{
	double result;
	double Channel_Temp;
	int sc_i;
	Channel_Temp = 0;

	for (sc_i = 0; sc_i < BlockData.Size.This; ++sc_i)
	{
		Channel_Temp += nsqr(1.0 * BlockData.Pointer[2][this_channel][sc_i]);
	}

	result = nlog2(Channel_Temp * Global.Codec_Block_size_recip) * SINGLE_HALF;
	return result;
}

void Shift_Codec_Blocks()
{
	int sc_i;
	MultiChannelCodecBlockPtr sc_p;
	int channel;
	double left_sample, right_sample;

	sc_p = BlockData.Pointer[0];
	BlockData.Pointer[0] = BlockData.Pointer[1];
	BlockData.Pointer[1] = BlockData.Pointer[2];
	BlockData.Pointer[2] = BlockData.Pointer[3];
	BlockData.Pointer[3] = sc_p;
	BlockData.Size.Prev = BlockData.Size.Last;
	BlockData.Size.Last = BlockData.Size.This;
	BlockData.Size.This = BlockData.Size.Next;
	BlockData.Size.Next = 0;

	if (BlockData.Size.This > 0)
	{
		Global.Codec_Block_size_recip = OneOver[BlockData.Size.This];
		++ Global.blocks_processed;
		Global.blocks_processed_recip = 1.0 / Global.blocks_processed;

		for (channel = 0; channel < Global.channels; ++channel)
		{
			BlockData.Channel_Log2_RMS[channel] = Channel_RMS(channel);
		}

		if (parameters.midside && (Global.channels == 2))
		{
			for (sc_i = 0; sc_i < BlockData.Size.This; ++sc_i)
			{
				left_sample = BlockData.Pointer[2][0][sc_i];
				right_sample = BlockData.Pointer[2][1][sc_i];
				BlockData.Pointer[2][2][sc_i] = nround(SINGLE_HALF * (left_sample + right_sample));
				BlockData.Pointer[2][3][sc_i] = nround(SINGLE_HALF * (left_sample - right_sample));
			}

			BlockData.Channel_Log2_RMS[2] = Channel_RMS(2);
			BlockData.Channel_Log2_RMS[3] = Channel_RMS(3);
		}
	}
	else
	{
		for (channel = 0; channel < Global.channels; ++channel)
		{
			BlockData.Channel_Log2_RMS[channel] = 0;
		}
	}
}
// initialization: todo
// finalization: todo Write ( Erroutput, UnitName, CRLF );
